/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2011 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on June 8, 2011, 12:49 AM
 */

#include <deque>

#include "u2d_DisplayContainer.h"

u2d_DisplayContainer::u2d_DisplayContainer(std::string _name /*= ""*/) :
enabled(true),
name(_name) {
}

/**
 * Notice that this will delete any u2d_DisplayObject that this
 * u2d_DisplayContainer contains. If you are using the object for something
 * else and you do not want it to get deleted it's best that you take it out
 * of the container and store it on your own.
 */
u2d_DisplayContainer::~u2d_DisplayContainer() {
    /* Shared pointer will fall out of scope and call delete on all the
     * pointers. This note is here to remember you exactly that in case you were
     * asking the following is not needed:
     * this->children.clear();
     * Shared pointer will fall out of scope and call delete on all the
     * pointers.
     */
}

/**
 * Adds a child u2d_DisplayObject to this container. From now on calling draw()
 * on the container will take care of calling draw() on the child also.
 * @param displayObject A pointer to a u2d_DisplayObject allocated on the heap
 * with new. Do not give in static u2d_DisplayObject (or any other derived
 * type) as this will cause delete to be called on them.
 * @remark This takes ownership of the pointer and will take care of
 * deallocating it.
 * @return A reference to this class used for named parameter idiom.
 */
u2d_DisplayContainer& u2d_DisplayContainer::addChild(u2d_DisplayObject *displayObject) {
    u2d_DisplayObjectPointer ptr(displayObject);
    this->children.push_back(ptr);
    return *this;
}

std::string u2d_DisplayContainer::getName() const {
    return this->name;
}

u2d_DisplayContainer& u2d_DisplayContainer::enable() {
    this->enabled = true;
    return *this;
}

u2d_DisplayContainer& u2d_DisplayContainer::disable() {
    this->enabled = false;
    return *this;
}

bool u2d_DisplayContainer::isEnabled() const {
    return enabled;
}

/**
 * Call draw() on all child DisplayObjects.
 * Classes inheriting from this one should call this method via
 * u2d_DisplayContainer::draw() so that draw gets called on child DisplayObjects.
 * Else they should implement this feature on their own.
 * @remark Draw methods gets called even on disabled DisplayObjects. It's the
 * object itself that should do nothing if it is not enabled.
 */
void u2d_DisplayContainer::draw() {
    if (this->isEnabled()) {
        std::deque<u2d_DisplayObjectPointer>::iterator childrenIterator = this->children.begin();
        while (childrenIterator != this->children.end()) {
            (*childrenIterator)->draw();
            ++childrenIterator;
        }
    }
}

