/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003-2011 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <math.h>
#include <vector>
#include <GL/gl.h>

#include <fltk/events.h>

#include "u2d_GLmesh.h"
#include "gl/u2d_DisplayObject.h"
#include "gl/u2d_DisplaySet.h"

#define BORDER 5
#define WHEEL_ZOOM_FACTOR -0.2
#define BUTTONPRESS_ZOOM_FACTOR 0.3

u2d_GLmesh::u2d_GLmesh(int X, int Y, int W, int H, const char *L) :
fltk::GlWindow(X, Y, W, H, L),
last_push_x(0),
last_push_y(0),
current_factor(1),
ref_minX(0),
ref_maxX(0),
ref_minY(0),
ref_maxY(0),
real_minX(0),
real_maxX(0),
real_minY(0),
real_maxY(0) {
    set_ref_sizes(-2, 2, -2, 2);
}

void u2d_GLmesh::draw() {
  if(!valid()) {
    /*projection
     *viewport
     *  w() and h() are the window dimensions
     *  valid() is turned on after FLTK draw() returns
     */
    /*
    ortho();
    */
    set_default_colors();
    set_default_antialiasing();
    glViewport(BORDER, BORDER, this->w() - 2 * BORDER, this->h() - 2 * BORDER);
  }
  /* draw */
  initialize();
  glClear(GL_COLOR_BUFFER_BIT);
  drawDisplayObjects();
  glFlush();
  

  //  swap_buffers();
}


int u2d_GLmesh::handle(int event) {
/* Don't call any OpenGL draw function from inside handle() */
  switch(event) {
    /* Those are events I want to to handle */
    case fltk::PUSH:
      if(fltk::event_button() == 1){ /* left button */
	last_push_x = fltk::event_x();
	last_push_y = fltk::event_y();
      }
      else if(fltk::event_button() == 2){ /* middle button */
	zoom(BUTTONPRESS_ZOOM_FACTOR, fltk::event_x(), fltk::event_y());
      }
      else if(fltk::event_button() == 3){ /* right button */
	zoom(-BUTTONPRESS_ZOOM_FACTOR, fltk::event_x(), fltk::event_y());
      }

      /* This say that we want to receive further related events too*/
    return 1;
  case fltk::DRAG:
      if(fltk::event_button() == 1){
	pan((double)(last_push_x - fltk::event_x()), (double)(last_push_y - fltk::event_y()));
	last_push_x = fltk::event_x();
	last_push_y = fltk::event_y();
      }
    return 1;
  case fltk::RELEASE:
    return 1;
  case fltk::MOUSEWHEEL:
    centered_zoom(((double)fltk::event_dy() * WHEEL_ZOOM_FACTOR));
    return 1;
  default:
    /* The base class here handle the others */
    return fltk::GlWindow::handle(event);
  }
}

void u2d_GLmesh::pan(double disp_x, double disp_y){
  double new_minX;
  double new_maxX;
  double new_minY;
  double new_maxY;
  /* Calculate the new reference sizes as old +/- the displacement
   * converted from window pixel units to OpenGL normalized units.
   * NOTE on +/- :
   * the fact that x has + and y has - is due to different convention
   * on the choice of the axis.
   * Event handler axis        |        OpenGL axis
   * (mouse coordinates)       |        (drawing coordinates)
   *                           |
   *   |                       |           ^
   * --+--------->             |           |
   *   |                       |           | 
   *   |                       |           |
   *   |                       |           |
   *   |                       |           |
   *   |                       |         --+---------->
   *   v                       |           |
   *
   * NOTE on normalization: 
   * We translate the reference sizes but normalize
   * on the real sizes, this is best behavior.
   */
  new_minX = ref_minX + disp_x * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  new_maxX = ref_maxX + disp_x * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  new_minY = ref_minY - disp_y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);
  new_maxY = ref_maxY - disp_y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);

#ifdef DEBUGGING_GL
  printf("Displacement X: %f Y: %f\n", disp_x / (this->w() - BORDER * 2), \
	 disp_y / (this->h() - BORDER * 2));
  printf("real_size are: minX: %g, maxX: %g, minY: %g, MaxY: %g\n", real_minX, real_maxX, real_minY, real_maxY);
#endif
  set_ref_sizes(new_minX, new_maxX, new_minY, new_maxY);
  redraw();
}

void u2d_GLmesh::centered_zoom(double delta_factor){
  int click_X = (this->w() - BORDER * 2) / 2;
  int click_Y = (this->h() - BORDER * 2) / 2;
  zoom(delta_factor, click_X, click_Y);
}

void u2d_GLmesh::zoom(double delta_factor, int click_X, int click_Y){
  /* We get delta_factor cause it is of how much we want to zoom and not
   * the final factor we want to go, 0.1 is indeed 10% more not 1/10 of
   * the normal factor.
   */
  double new_minX;
  double new_maxX;
  double new_minY;
  double new_maxY;
  double length_X;
  double length_Y;
  double center_X;
  double center_Y;
  double visible_area;
  double zoomed_area;
  double aspect_ratio;
  /* We do an area zoom, not a length zoom */
  visible_area = fabs(real_maxX - real_minX) * fabs(real_maxY - real_minY);
  zoomed_area = visible_area * ( 1 - delta_factor );
#ifdef DEBUGGING_GL
  printf("Delta factor: %f Visible area: %f, zoomed_area: %f\n",\
	 delta_factor, visible_area, zoomed_area);
#endif
  aspect_ratio =  fabs(real_maxX - real_minX) / fabs(real_maxY - real_minY);
  length_X = sqrt(zoomed_area * aspect_ratio);
  length_Y = length_X / aspect_ratio;
#ifdef DEBUGGING_GL
  printf("length_X: %f length_Y: %f\n", length_X, length_Y);
#endif
  /* NOTE on +/-:
   * see pan() for explanation why we use +/-
   */
  center_X = real_minX + click_X * fabs(real_maxX - real_minX) / (this->w() - BORDER * 2);
  center_Y = real_maxY - click_Y * fabs(real_maxY - real_minY) / (this->h() - BORDER * 2);
#ifdef DEBUGGING_GL
  printf("center_X: %f center_Y: %f\n", center_X, center_Y);
#endif
  new_minX = center_X - length_X / 2;
  new_maxX = center_X + length_X / 2;
  new_minY = center_Y - length_Y / 2;
  new_maxY = center_Y + length_Y / 2;
  set_ref_sizes(new_minX, new_maxX, new_minY, new_maxY);
  redraw();
}


/**
 * Call glCallList on all display_set s.
 * If a given display_set has not a valid display list
 * it is just skipped.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::drawDisplayObjects(){
    std::vector<boost::shared_ptr<u2d_DisplayObject> >::iterator iter;
    for(iter = displayObjects.begin(); iter != displayObjects.end(); iter++) {
            (*iter)->draw();
    }
}

/**
 * Pushes a displaySet in the list of displaySets to draw. This method also
 * gets the displaySet ready and not modifiable.
 * @param _displaySet The display set to add.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::addDisplayObject(u2d_DisplayObject *_displayObject) {
    u2d_DisplayObjectPointer ptr(_displayObject);
    this->displayObjects.push_back(ptr);
}

/**
 * Adds all displaySets in the given vector to those to draw. This method
 * also gets all the given display sets ready and so not modifiable.
 * @remark Notice that this function will empty the given vector since
 * the class will take ownership of the pointers. The memory area pointed
 * by the vectors will be automatically freed by the class when the pointers
 * fall out of scope.
 * @param _displaySet The vector with the dispalySets to add.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::addDisplayObject(std::vector<u2d_DisplayObject *> _displayObject) {
    std::vector<u2d_DisplayObject *>::iterator iter;
    for(iter = _displayObject.begin(); iter != _displayObject.end(); iter++) {
        /* Note, smart pointer conversion must be done _explicitely_ */
        // I'm quite sure that this local named will be optimized away, moreover
        // boost best practices say to avoid temporaries of shared pointers
        u2d_DisplayObjectPointer ptr(*iter); 
        this->displayObjects.push_back(ptr);
    }
    _displayObject.clear();
}

/**
 * Adds all displaySets in the given vector to those to draw. This method
 * also gets all the given display sets ready and so not modifiable.
 * @remark Notice that this function will empty the given vector since
 * the class will take ownership of the pointers. The memory area pointed
 * by the vectors will be automatically freed by the class when the pointers
 * fall out of scope.
 * @param _displaySet The vector with the dispalySets to add.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::addDisplayObject(std::vector<u2d_DisplaySet *> _displaySet) {
    std::vector<u2d_DisplaySet *>::iterator iter;
    for(iter = _displaySet.begin(); iter != _displaySet.end(); iter++) {
        /* Note, smart pointer conversion must be done _explicitely_ */
        // I'm quite sure that this local named will be optimized away, moreover
        // boost best practices say to avoid temporaries of shared pointers
        u2d_DisplayObjectPointer ptr(*iter);
        this->displayObjects.push_back(ptr);
    }
    _displaySet.clear();
}


/**
 * Removes the first displaySet whose name matches the given name. The display
 * set is also deleted and it's memory deallocated.
 * @param name The name of the displaySets to remove.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::removeDisplayObject(string name) {
    std::vector<boost::shared_ptr<u2d_DisplayObject> >::iterator it  = this->displayObjects.begin();
    while(it != this->displayObjects.end()) {
        if(name.compare((*it)->getName()) == 0 ) {
            /* Regenerate the iterator because erase invalidates it*/
            it = this->displayObjects.erase(it);
            /* We bail out. N.B. The regenerated iterator is not used.*/
            return;
        } else {
            it++;
        }
    }
    return;
}

/**
 * Removes all displaySets.
 * @remark Notice that all DisplaySets contained in this class will be deleted.
 * If you wish to keep some for later use, you have to do that manually.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_GLmesh::clearDisplayObjects() {
    this->displayObjects.clear();
}


void u2d_GLmesh::set_ref_sizes(double minX, double maxX, double minY, double maxY){
  /* This is a fix for:
   * When autoscaling the gl window, and only a one dimensional element,
   * like an horizontal or vertical line, is given the scale can't be
   * guessed correctly, that's because we do not have a reference
   * dimension for both X and Y so one of those will be 0 actually
   * preventing correct drawing.
   */

  if((minX != 0 || maxX != 0) && (minY == 0 && maxY == 0)){
#ifdef DEBUGGING_GL
    printf("X size overriding Y\n");
#endif
    minY = minX;
    maxY = maxX;
  }
  else if((minY != 0 || maxY != 0) && (minX == 0 && maxX == 0)){
#ifdef DEBUGGING_GL
    printf("Y size overriding X\n");
#endif
    minX = maxY;
    maxX = maxY;
  }
#ifdef DEBUGGING_GL
  printf("ref_size sets: minX: %g, maxX: %g, minY: %g, MaxY: %g\n", minX, maxX, minY, maxY);
#endif
  ref_minX = minX;
  ref_maxX = maxX;
  ref_minY = minY;
  ref_maxY = maxY;
}

void u2d_GLmesh::get_ref_sizes(double *minX, double *maxX, double *minY, double *maxY){

  *minX = ref_minX;
  *maxX = ref_maxX;
  *minY = ref_minY;
  *maxY = ref_maxY;

}

void u2d_GLmesh::get_real_sizes(double *minX, double *maxX, double *minY, double *maxY){

  *minX = real_minX;
  *maxX = real_maxX;
  *minY = real_minY;
  *maxY = real_maxY;

}

void u2d_GLmesh::initialize(){

  float factor;
  double view_minX;
  double view_maxX;
  double view_minY;
  double view_maxY;
  double center;
 
  /* This keep aspect ratio */

  double deltaY = ref_maxY - ref_minY;
  double deltaX = ref_maxX - ref_minX;
  if((deltaX / (double)this->w()) < (deltaY / (double)this->h())) {
    factor = (deltaY / (double)this->h()) / (deltaX / (double)this->w());
    center = (ref_maxX + ref_minX) / 2.0;
    view_minY = ref_minY;
    view_maxY = ref_maxY;
    view_minX = center - (center - ref_minX)*factor;
    view_maxX = center + (ref_maxX - center)*factor;
  }
  else {
    factor = (deltaX / (double)this->w()) / (deltaY / (double)this->h());
    center = (ref_maxY + ref_minY) / 2.0;
    view_minY = center - (center - ref_minY)*factor;
    view_maxY = center + (ref_maxY - center)*factor;
    view_minX = ref_minX;
    view_maxX = ref_maxX;
  }

  /* We take as reference the smaller area drawable */
  /* This algo does not keep aspect ratio */
  /*
  if(this->h() >= this->w()){
    factor = (double)this->h() / (double)this->w();
    center = (ref_maxY + ref_minY) / 2.0;
    view_minY = center - (center - ref_minY)*factor;
    view_maxY = center + (ref_maxY - center)*factor;
    view_minX = ref_minX;
    view_maxX = ref_maxX;
  }
  else{
    factor = (double)this->w() / (double)this->h();
    center = (ref_maxX + ref_minX) / 2.0;
    view_minY = ref_minY;
    view_maxY = ref_maxY;
    view_minX = center - (center - ref_minX)*factor;
    view_maxX = center + (ref_maxX - center)*factor;
  }
  */

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(view_minX * 1.00, view_maxX * 1.00, view_minY * 1.00, view_maxY * 1.00, -1, 1 );

  /* Set the real coordinates we are using */
  /* We have real sizes to improve user experience while panning
   * and zooming, here below description of the bug fixed with real
   * sizes addition.
   * The fact is that due to one size being bigger than his actual
   * ref_size (to keep aspect ratio) mouse movement is not consistent
   * with actual pan, cause the figure "slide" under the mouse pointer
   * when panning along the biggest dimension.
   * Proposed fix:
   * Add a real_minX and so on, they get set in set_ref_sizes and use
   * them instead of ref_minX and so on.
   */
  real_minX = view_minX;
  real_maxX = view_maxX;
  real_minY = view_minY;
  real_maxY = view_maxY;
}

void u2d_GLmesh::set_default_colors(){
  /* Set up colors and line width */
  glClearColor(0.0, 0.0, 0.0, 0.0); //black
  glClear(GL_COLOR_BUFFER_BIT); // the background
  glColor3f(1.0, 1.0, 1.0);  //white
  glLineWidth((GLfloat)1); // 1 width for lines
}

void u2d_GLmesh::set_default_antialiasing(){
  /* The next 3 lines make openGL lines anti-aliased and good looking,
   * removing one will not do the trick anymore, look like they
   * work together
   */
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}
